<?php

declare(strict_types=1);

namespace Drupal\poll_rest\Plugin\rest\resource;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\poll\Entity\Poll;
use Drupal\poll\PollInterface;
use Drupal\poll\PollVoteStorageInterface;
use Drupal\rest\ModifiedResourceResponse;
use Drupal\rest\Plugin\ResourceBase;
use Drupal\rest\ResourceResponse;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Represents poll records as resources.
 *
 * @RestResource(
 *   id = "poll_vote_resource",
 *   label = @Translation("Poll vote resource"),
 *   uri_paths = {
 *     "canonical" = "/api/v1/poll/{pid}",
 *     "create" = "/api/v1/poll/{pid}/vote"
 *   }
 * )
 */
final class PollVoteResource extends ResourceBase {

  /**
   * The storage interface for poll votes.
   *
   * @var \Drupal\poll\PollVoteStorageInterface
   */
  protected $pollVoteStorage;

  /**
   * Stores currentUser.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * Stores request.
   *
   * @var Symfony\Component\HttpFoundation\Request
   */
  protected $currentRequest;

  /**
   * The time service interface.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * Constructs a Drupal\rest\Plugin\ResourceBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param array $serializer_formats
   *   The available serialization formats.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\poll\PollVoteStorageInterface $pollVoteStorage
   *   The poll storage class.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user instance.
   * @param \Symfony\Component\HttpFoundation\Request $current_request
   *   The current request.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    array $serializer_formats,
    LoggerInterface $logger,
    PollVoteStorageInterface $pollVoteStorage,
    AccountProxyInterface $current_user,
    Request $current_request,
    TimeInterface $time,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $serializer_formats, $logger);
    $this->currentUser = $current_user;
    $this->pollVoteStorage = $pollVoteStorage;
    $this->currentRequest = $current_request;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->getParameter('serializer.formats'),
      $container->get('logger.factory')->get('poll_rest'),
      $container->get('poll_vote.storage'),
      $container->get('current_user'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('datetime.time')
    );
  }

  /**
   * Responds to GET requests.
   *
   * @return \Drupal\rest\ResourceResponse
   *   The HTTP response object.
   *
   * @throws \Symfony\Component\HttpKernel\Exception\HttpException
   *   Throws exception expected.
   */
  public function get($pid): ResourceResponse {
    // The current user must have access to vote.
    if (!$this->currentUser->hasPermission('access polls')) {
      throw new AccessDeniedHttpException('Current user must have access to vote.');
    }
    $poll = $this->getPoll($pid);

    if (empty($this->currentRequest->query->get('query'))) {
      $this->logger->notice('params query is empty');
      throw new BadRequestHttpException("params query is empty");
    }

    $query = $this->currentRequest->query->get('query');
    $allowed_values = ['votes', 'choices'];
    if (stripos(json_encode($allowed_values), $query) !== FALSE) {
      if ($query === 'choices') {
        if ($this->isVotingAllowed($poll)) {
          $options = $poll->getOptions();
          $data = $this->processOptions($poll, $options);
        }
      }
      else {
        if ($this->currentUser->hasPermission('view poll results')) {
          $data = $this->getPollResults($poll);
        }
        else {
          throw new AccessDeniedHttpException("User does not have permissions to view poll results.");
        }
      }
    }
    else {
      $this->logger->notice('params query must be choices or votes');
      throw new NotFoundHttpException("params query must be choices or votes");
    }

    $response = new ResourceResponse($data);
    $response->addCacheableDependency($data);

    // In order to generate fresh result every time (without clearing
    // the cache), you need to invalidate the cache.
    $this->logger->notice("GET {$query} of poll: {$pid}.");

    // @todo: Implements cache.
    // $build = [
    //   '#cache' => [
    //     'contexts' => ['url.query_args'],
    //     'tags' => $this->getCacheTags($pid, $query),
    //   ],
    // ];

    // $response->addCacheableDependency(CacheableMetadata::createFromRenderArray($build));
    return $response;
  }

  /**
   * Response to the POST request.
   *
   * @param int $pid
   *   The poll ID.
   * @param array $data
   *   The data array.
   *
   * @return \Drupal\rest\ModifiedResourceResponse
   *   The modified resource response.
   *
   * @throws \Symfony\Component\HttpKernel\Exception\HttpException
   *   Throws an exception if there is an error.
   */
  public function post($pid, array $data): ModifiedResourceResponse {
    $poll = $this->getPoll($pid);
    $this->isVotingAllowed($poll);

    // Validate the choice.
    $choices = $poll->getOptions();
    if (!isset($choices[$data['chid']])) {
      return new ModifiedResourceResponse(['message' => 'Invalid Choice'], 400);
    }

    $options = [];
    $options['chid'] = $data['chid'];
    $options['uid'] = $this->currentUser->id();
    $options['pid'] = $poll->id();
    $options['hostname'] = $this->currentRequest->getClientIp();
    $options['timestamp'] = $this->time->getRequestTime();

    try {
      $url = $poll->toUrl()->toString();
      $this->pollVoteStorage->saveVote($options);
    }
    catch (\Exception $e) {
      throw new BadRequestHttpException("Vote could not be saved: {$e->getMessage()}");
    }

    $this->logger->notice('%user\'s vote has been recorded.', [
      '%user' => $this->currentUser->id(),
    ]);

    // $tags = [
    // "poll-votes:{$pid}",
    // ];
    // $this->cacheTagsInvalidator->invalidateTags($tags);
    $responseData = ['error' => FALSE];
    return new ModifiedResourceResponse($responseData, 201);
  }

  /**
   * Response to the DELETE request.
   *
   * @param int $pid
   *   The poll ID.
   *
   * @return \Drupal\rest\ModifiedResourceResponse
   *   The response indicating the result of the delete operation.
   *
   * @throws \Symfony\Component\HttpKernel\Exception\HttpException
   *   Throws an exception if there is an error.
   */
  public function delete($pid): ModifiedResourceResponse {
    $poll = $this->getPoll($pid);

    if (!$this->isCancelAllowed($poll)) {
      throw new AccessDeniedHttpException("User has not allowed to cancel vote for poll {$pid}.");
    }

    try {
      $this->pollVoteStorage->cancelVote($poll, $this->currentUser);
    }
    catch (\Exception $e) {
      throw new BadRequestHttpException("Vote could not be cancel: {$e->getMessage()}");
    }

    $this->logger->notice('%user\'s vote in Poll #%poll deleted.', [
      '%user' => $this->currentUser->id(),
      '%poll' => $pid,
    ]);

    // $tags = [
    // "poll-votes:{$pid}",
    // ];
    // $this->cacheTagsInvalidator->invalidateTags($tags);
    $responseData = ['error' => FALSE];
    return new ModifiedResourceResponse($responseData, 200);
  }

  /**
   * Returns poll object.
   *
   * @param int $pid
   *   The poll ID.
   *
   * @return \Drupal\poll\PollInterface
   *   The Poll object.
   *
   * @throws \Symfony\Component\HttpKernel\Exception\HttpException
   *   Throws an exception if the poll cannot be found.
   */
  protected function getPoll($pid): PollInterface {
    $poll = Poll::load($pid);

    // Check if poll found.
    if (is_null($poll)) {
      throw new NotFoundHttpException("Poll {$pid} not found.");
    }

    // Check if poll is open.
    if (!$poll->isOpen()) {
      throw new AccessDeniedHttpException("Poll {$pid} is inactive.");
    }

    return $poll;
  }

  /**
   * Returns cacheTags.
   *
   * @param string $pid
   *   The poll id.
   * @param string $query
   *   The $query string.
   *
   * @return array
   *   data array.
   */
  protected function getCacheTags(string $pid, string $query): array {

    if ($query === 'choices') {
      $tags = [
        "poll:{$pid}",
      ];
    }
    else {
      $tags = [
        "poll-votes:{$pid}",
      ];
    }
    return $tags;
  }

  /**
   * Checks if the current user is allowed to vote or cancel on the given poll.
   *
   * @param \Drupal\poll\PollInterface $poll
   *   The poll to check.
   *
   * @return bool
   *   True if the user can vote, false otherwise.
   */
  protected function isVotingAllowed(PollInterface $poll) {
    // The current user must have access to vote.
    if (!$this->currentUser->hasPermission('access polls')) {
      throw new AccessDeniedHttpException('Current user must have access to vote.');
    }

    // If the poll is closed, we're not allowed to vote.
    if ($poll->isClosed()) {
      throw new AccessDeniedHttpException('Poll is closed');
    }

    // @todo If we are anonymous the poll is not allowed.
    if ($this->currentUser->isAnonymous()) {
      throw new AccessDeniedHttpException('User is anonymous');
    }

    // If the user hasn't voted yet, voting is allowed.
    if ($poll->hasUserVoted() === FALSE) {
      return TRUE;
    }

    throw new AccessDeniedHttpException('User is not allowed');
  }

  /**
   * Checks if the current user is allowed to cancel on the given poll.
   *
   * @param \Drupal\poll\PollInterface $poll
   *   A poll.
   *
   * @return bool
   *   TRUE if the user can cancel.
   */
  protected function isCancelAllowed(PollInterface $poll): bool {
    // Allow access if the user has voted.
    return $poll->hasUserVoted()
      // And the poll allows to cancel votes.
      && $poll->getCancelVoteAllow()
      // And the user has the cancel own vote permission.
      && $this->currentUser->hasPermission('cancel own vote')
      // And the user is authenticated or the session contains the voted flag.
      && ($this->currentUser->isAuthenticated())
      // And poll is open.
      && $poll->isOpen();
  }

  /**
   * Results of a poll.
   *
   * @param \Drupal\poll\PollInterface $poll
   *   The poll entity.
   *
   * @return array
   *   data array.
   */
  protected function getPollResults(PollInterface $poll): array {
    $options = $poll->getOptions();
    $poll_results = [];
    $votes = $poll->getVotes();
    $total_votes = $this->pollVoteStorage->getTotalVotes($poll);
    $user_vote = $this->pollVoteStorage->getUserVote($poll);
    switch ($poll->getVotesOrderType()) {
      case PollInterface::VOTES_ORDER_COUNT_ASC:
        asort($votes);
        break;

      case PollInterface::VOTES_ORDER_COUNT_DESC:
        arsort($votes);
        break;
    }
    foreach ($votes as $pid => $vote) {
      $percentage = round($vote * 100 / max($total_votes, 1));
      $display_votes = ' (' . $this->getStringTranslation()
        ->formatPlural($vote, '1 vote', '@count votes') . ')';
      $current_vote = $user_vote ? $user_vote['chid'] : FALSE;
      $poll_results[] = [
        'choice' => $options[$pid],
        'display_value' => $this->t('@percentage%', ['@percentage' => $percentage]) . $display_votes,
        'value' => $vote,
        'percentage' => $percentage,
        'is_current_selection' => (int) $current_vote === $pid,
      ];
    }

    $output = [
      'raw_question' => $poll->label(),
      'total_votes' => $total_votes,
      'pid' => $poll->id(),
      'user_vote' => $user_vote['chid'] ?? NULL,
      'results' => $poll_results,
    ];

    return $output;
  }

  /**
   * {@inheritdoc}
   */
  protected function processOptions(PollInterface $poll, array $values) {
    $data = [];

    foreach ($values as $chid => $name) {
      // $action = '/api/v1/poll/' . $poll->id(). '/vote';
      $data[] = [
        'chid' => strval($chid),
        'name' => $name,
        // 'action' => $action,
      ];
    }

    $output = [
      'raw_question' => $poll->label(),
      'pid' => $poll->id(),
      'user_vote' => $user_vote['chid'] ?? NULL,
      'choices' => $data,
    ];

    return $output;
  }

}
